"""
File input/output functions for molecular geometry files.

Can support XYZ, COLUMBUS and Z-matrix formats. Input and output both
require an open file to support multiple geometries. Only 3D geometries
are currently supported.
"""
import numpy as np
import gimbal.constants as con
import gimbal.displace as displace
import gimbal.measure as measure


def read_xyz(infile, units='ang', hasvec=False, hascom=False):
    """Reads input file in XYZ format.

    XYZ files are in the format::

        natm
        comment
        A X1 Y1 Z1 [Vx1 Vy1 Vz1]
        B X2 Y2 Z2 [Vx2 Vy2 Vz2]
        ...

    where natm is the number of atoms, comment is a comment line, A and B
    are atomic labels and X, Y and Z are cartesian coordinates (in
    Angstroms). The vectors Vx, Vy and Vz are optional and will only be
    read if hasvec = True.

    Due to the preceding number of atoms, multiple XYZ format geometries
    can easily be read from a single file.

    Parameters
    ----------
    infile : file
        The open input file.
    units : str, optional
        The units of length of the cartesian coordinates. Default is
        Angstroms.
    hasvec : bool, optional
        Specifies if a vector should be read from the input file. If
        False (default), vec is a zero array.
    hascom : bool, optional
        Specifies if a comment line should be read from the input file. If
        False (default), comment is a blank string.

    Returns
    -------
    elem : (N,) ndarray
        The atomic symbols.
    xyz : (N, 3) ndarray
        The atomic cartesian coordinates.
    vec : (N, 3) ndarray
        The vector cartesian coordinates.
    comment : str
        The comment line.

    Raises
    ------
    IOError
        When geometry is not in the correct format (natm can't be read).
    """
    try:
        natm = int(infile.readline())
    except ValueError:
        raise IOError('geometry not in XYZ format.')
    if hascom:
        comment = infile.readline().strip()
    else:
        infile.readline()
        comment = ''
    data = np.array([infile.readline().split() for i in range(natm)])
    elem = data[:, 0]
    xyz = data[:, 1:4].astype(float) * con.conv(units, 'ang')
    if hasvec:
        vec = data[:, 4:7].astype(float)
        if vec.size == 0:
            vec = np.zeros_like(xyz)
    else:
        vec = None
    return elem, xyz, vec, comment


def read_col(infile, units='bohr', hasvec=False, hascom=False):
    """Reads input file in COLUMBUS format.

    COLUMBUS geometry files are in the format::

        A nA X1 Y1 Z1 mA
        B nB X2 Y2 Z2 mB
        ...

    where A and B are atomic labels, nA and nB are corresponding atomic
    numbers, mA and mB are corresponding atomic masses and X, Y and Z
    are cartesian coordinates (in Bohrs).

    COLUMBUS geometry files do not provide the number of atoms in each
    geometry. A comment line (or blank line) must be used to separate
    molecules.

    For the time being, vector input is not supported for the
    COLUMBUS file format.

    Parameters
    ----------
    infile : file
        The open input file.
    units : str, optional
        The units of length of the cartesian coordinates. Default is
        Bohr.
    hasvec : bool, optional
        Specifies if a vector should be read from the input file. If
        False (default), vec is a zero array.
    hascom : bool, optional
        Specifies if a comment line should be read from the input file. If
        False (default), comment is a blank string.

    Returns
    -------
    elem : (N,) ndarray
        The atomic symbols.
    xyz : (N, 3) ndarray
        The atomic cartesian coordinates.
    vec : (N, 3) ndarray
        The vector cartesian coordinates.
    comment : str
        The comment line.
    """
    if hascom:
        comment = infile.readline().strip()
    else:
        comment = ''
    data = np.empty((0, 6), dtype=str)
    while True:
        pos = infile.tell()
        line = np.array(infile.readline().split())
        try:
            # catch comment line or end-of-file
            line[1:].astype(float)
            data = np.vstack((data, line))
        except (ValueError, IndexError):
            if len(data) < 1:
                raise IOError('geometry not in COLUMBUS format.')
            else:
                # roll back one line before break
                infile.seek(pos)
                break
    elem = data[:, 0]
    xyz = data[:, 2:5].astype(float) * con.conv(units, 'ang')
    if hasvec:
        vec = np.zeros_like(xyz)
    else:
        vec = None
    return elem, xyz, vec, comment


def read_gdat(infile, units='bohr', hasvec=False, hascom=False):
    """Reads input file in FMS90 Geometry.dat format.

    Geometry.dat files are in the format::

        comment
        natm
        A X1 Y1 Z1
        B X2 Y2 Z2
        ...
        Vx1 Vy1 Vz1
        Vx2 Vy2 Vz2
        ...

    where comment is a comment line, natm is the number of atoms, A and B
    are atomic labels, X, Y and Z are cartesian coordinates and Vq are
    vectors for cartesian coordinates q. The vectors are only read if
    hasvec = True.

    Parameters
    ----------
    infile : file
        The open input file.
    units : str, optional
        The units of length of the cartesian coordinates. Default is
        Bohr.
    hasvec : bool, optional
        Specifies if a vector should be read from the input file. If
        False (default), vec is a zero array.
    hascom : bool, optional
        Specifies if a comment line should be read from the input file. If
        False (default), comment is a blank string.

    Returns
    -------
    elem : (N,) ndarray
        The atomic symbols.
    xyz : (N, 3) ndarray
        The atomic cartesian coordinates.
    vec : (N, 3) ndarray
        The vector cartesian coordinates.
    comment : str
        The comment line.

    Raises
    ------
    IOError
        When geometry is not in the correct format (natm can't be read).
    """
    if hascom:
        comment = infile.readline().strip()
    else:
        infile.readline()
        comment = ''
    try:
        natm = int(infile.readline())
    except ValueError:
        raise IOError('geometry not in Geometry.dat format')
    data = np.array([infile.readline().split() for i in range(natm)])
    elem = data[:, 0]
    xyz = data[:, 1:].astype(float) * con.conv(units, 'ang')
    if hasvec:
        vec = np.array([infile.readline().split() for i in range(natm)],
                       dtype=float)
    else:
        vec = None
    return elem, xyz, vec, comment


def read_zmt(infile, units='ang', hasvec=False, hascom=False):
    """Reads input file in Z-matrix format.

    Z-matrix files are in the format::

        A
        B 1 R1
        C indR2 R2 indA2 A2
        D indR3 R3 indA3 A3 indT3 T3
        E indR4 R4 indA4 A4 indT4 T4
        ...

    where A, B, C, D, E are atomic labels, indR, indA, indT are reference
    atom indices, R are bond lengths (in Angstroms), A are bond angles (in
    degrees) and T are dihedral angles (in degrees). For example, E is a
    distance R from atom indR with an E-indR-indA angle of A and an
    E-indR-indA-indT dihedral angle of T. Alternatively, values can be
    assigned to a list of variables after the Z-matrix (preceded by a blank
    line).

    Although the number of atoms is not provided, the unique format of the
    first atom allows multiple geometries to be read without separation
    by a comment line.

    For the time being, vector input is not supported for the
    Z-matrix file format.

    Parameters
    ----------
    infile : file
        The open input file.
    units : str, optional
        The units of length of the cartesian coordinates. Default is
        Angstroms.
    hasvec : bool, optional
        Specifies if a vector should be read from the input file. If
        False (default), vec is a zero array.
    hascom : bool, optional
        Specifies if a comment line should be read from the input file. If
        False (default), comment is a blank string.

    Returns
    -------
    elem : (N,) ndarray
        The atomic symbols.
    xyz : (N, 3) ndarray
        The atomic cartesian coordinates.
    vec : (N, 3) ndarray
        The vector cartesian coordinates.
    comment : str
        The comment line.

    Raises
    ------
    IOError
        When geometry is not in the correct format (no atoms found).
    """
    if hascom:
        comment = infile.readline().strip()
    else:
        comment = ''
    data = []
    vlist = dict()
    while True:
        pos = infile.tell()
        line = infile.readline()
        split = line.split()
        if line == '':
            # end-of-file
            break
        elif len(split) == 1 and len(data) > 0:
            # roll back one line before break
            infile.seek(pos)
            break
        elif split == []:
            # blank line before variable assignment
            continue
        elif split[0] in con.sym:
            data.append(split)
        elif split[1] == '=' and len(split) == 3:
            vlist[split[0]] = float(split[2])
        else:
            # assume it's a comment line and roll back
            infile.seek(pos)
            break

    natm = len(data)
    if natm < 1:
        raise IOError('geometry not in Z-matrix format.')
    elem = np.array([line[0] for line in data])
    xyz = np.zeros((natm, 3))
    for i in range(natm):
        if i == 0:
            # leave first molecule at origin
            continue
        elif i == 1:
            # move along z-axis by R
            xyz = displace.translate(xyz, _valvar(data[1][2], vlist),
                                     [0, 0, 1], ind=1)
        elif i == 2:
            indR = int(data[2][1]) - 1
            indA = int(data[2][3]) - 1
            xyz[2] = xyz[indR]
            # move from indR towards indA by R
            xyz = displace.translate(xyz, _valvar(data[2][2], vlist),
                                     xyz[indA]-xyz[indR], ind=2)
            # rotate into xz-plane by A
            xyz = displace.rotate(xyz, _valvar(data[2][4], vlist), [0, 1, 0],
                                  ind=2, origin=xyz[indR], units='deg')
        else:
            indR = int(data[i][1]) - 1
            indA = int(data[i][3]) - 1
            indT = int(data[i][5]) - 1
            xyz[i] = xyz[indR]
            # move from indR towards indA by R
            xyz = displace.translate(xyz, _valvar(data[i][2], vlist),
                                     xyz[indA]-xyz[indR], ind=i)
            # rotate about (indT-indA)x(indR-indA) by A
            xyz = displace.rotate(xyz, _valvar(data[i][4], vlist),
                                  np.cross(xyz[indT]-xyz[indA],
                                           xyz[indR]-xyz[indA]),
                                  ind=i, origin=xyz[indR], units='deg')
            # rotate about indR-indA by T
            xyz = displace.rotate(xyz, _valvar(data[i][6], vlist),
                                  xyz[indR]-xyz[indA],
                                  ind=i, origin=xyz[indR], units='deg')

    xyz = displace.centre_mass(elem, xyz) * con.conv(units, 'ang')
    if hasvec:
        vec = np.zeros_like(xyz)
    else:
        vec = None
    return elem, xyz, vec, comment


def read_traj(infile, units='bohr', hasvec=False, hascom=False,
              elem=None, time=None, autocom=False):
    """Reads input file in FMS/nomad trajectory format

    trajectory files are in the format::

        T1 X1 Y1 Z1 X2 Y2 ... Vx1 Vy1 Vz1 Vx2 Vy2 ... G Re(A) Im(A) |A| S
        T2 X1 Y1 Z1 X2 Y2 ... Vx1 Vy1 Vz1 Vx2 Vy2 ... G Re(A) Im(A) |A| S
        ...

    where T is the time, Vq are the vectors (momenta) for cartesian
    coordinates q, G is the phase, A is the amplitude and S is the state
    label. The vectors are only read if hasvec = True.

    Trajectory files do not contain atomic labels. If not provided, they are
    set to dummy atoms which may affect calculations involving atomic
    properties. A time should be provided, otherwise the first geometry
    in the file is used.

    Parameters
    ----------
    infile : file
        The open input file.
    units : str, optional
        The units of length of the cartesian coordinates. Default is
        Bohr.
    hasvec : bool, optional
        Specifies if a vector should be read from the input file. If
        False (default), vec is a zero array.
    hascom : bool, optional
        Specifies if a comment line should be read from the input file. If
        False (default), comment is a blank string.
    elem : (N,) array_like, optional
        A list of the atomic symbols. If elem is None (default), the
        symbols will all be set to 'X'.
    time : float, optional
        The desired time of the trajectory. If time is None (default), the
        first geometry in the buffer is parsed. Otherwise, the file is
        read until the specified time is found.
    autocom : bool, optional
        Specifies if a comment line should be automatically generated with
        time, state and squared amplitude. Default is False.

    Returns
    -------
    elem : (N,) ndarray
        The atomic symbols.
    xyz : (N, 3) ndarray
        The atomic cartesian coordinates.
    vec : (N, 3) ndarray
        The vector cartesian coordinates.
    comment : str
        The comment line.

    Raises
    ------
    IOError
        When the geometry is not in the correct format (empty line or
        incorrect number of columns).
    """
    if hascom:
        comment = infile.readline().strip()
    else:
        comment = ''
    if time is None:
        rawline = infile.readline().split()
        if rawline == []:
            raise IOError('empty line provided')
        elif 'Time' in rawline:
            line = np.array(infile.readline().split(), dtype=float)
        else:
            line = np.array(rawline, dtype=float)
    else:
        alldata = np.array([line.split() for line in infile.readlines()
                            if 'Time' not in line], dtype=float)
        line = alldata[np.isclose(alldata[:,0], time)][0]
    natm = len(line) // 6 - 1
    if natm < 1 or len(line) % 6 != 0:
        raise IOError('geometry not in trajectory format.')
    if elem is None:
        elem = np.array(['X'] * natm)
    xyz = line[1:3*natm+1].reshape(natm, 3) * con.conv(units,'ang')
    if hasvec:
        vec = line[3*natm+1:6*natm+1].reshape(natm, 3)
    else:
        vec = None
    if autocom:
        fmt = 't={:8.2f}, state={:4d}, a^2={:10.4f}'
        comment += fmt.format(line[0], int(line[-1]), line[-2])
    return elem, xyz, vec, comment


def read_auto(infile, hasvec=False, hascom=False, **kwargs):
    """Reads a molecular geometry file and determines the format.

    Parameters
    ----------
    infile : file
        The open input file.
    hasvec : bool, optional
        Specifies if a vector should be read from the input file.
    hascom : bool, optional
        Specifies if a comment line should be read from the input file.
    kwargs : dict, optional
        Additional keyword arguments for the read functions.

    Returns
    -------
    elem : (N,) ndarray
        The atomic symbols.
    xyz : (N, 3) ndarray
        The atomic cartesian coordinates.
    vec : (N, 3) ndarray
        The vector cartesian coordinates.
    comment : str
        The comment line.

    Raises
    ------
    IOError
        When the geometry format is not recognized by the input parser.
    """
    kwargs.update(dict(hascom=hascom, hasvec=hasvec))
    pos = infile.tell()
    contents = infile.readlines()
    infile.seek(pos)
    nlines = min(len(contents), 4)
    fmt = []
    for i in range(nlines):
        line = contents[i].split()
        fmt.append(''.join([_get_type(l) for l in line]))

    if nlines == 0:
        raise IOError('end of file')
    elif nlines == 1:
        if hascom:
            raise IOError('cannot have comment with single line file')
        else:
            if fmt[0] == 's':
                return read_zmt(infile, **kwargs)
            elif fmt[0].replace('i', 'f') == 'sfffff':
                return read_col(infile, **kwargs)
            elif 'ffffffffffff' in fmt[0].replace('i', 'f'):
                return read_traj(infile, **kwargs)
            else:
                raise IOError('single line input in unrecognized format')
    elif nlines == 2 and hascom:
        if fmt[1] == 's':
            return read_zmt(infile, **kwargs)
        elif fmt[1].replace('i', 'f') == 'sfffff':
            return read_col(infile, **kwargs)
        elif 'ffffffffffff' in fmt[1].replace('i', 'f'):
            return read_traj(infile, **kwargs)
        else:
            raise IOError('single line input in unrecognized format')
    else:
        if hascom:
            if fmt[1] == 's' and fmt[2] in ['sis', 'sif', 'sii']:
                return read_zmt(infile, **kwargs)
            elif [f.replace('i', 'f') for f in fmt[1:3]] == ['sfffff', 'sfffff']:
                return read_col(infile, **kwargs)
        else:
            if fmt[0] == 's' and fmt[1] in ['sis', 'sif', 'sii']:
                return read_zmt(infile, **kwargs)
            elif [f.replace('i', 'f') for f in fmt[:2]] == ['sfffff', 'sfffff']:
                return read_col(infile, **kwargs)

        if ('ffffffffffff' in fmt[0].replace('i', 'f') or
            'ffffffffffff' in fmt[1].replace('i', 'f')):
            return read_traj(infile, **kwargs)
        elif nlines > 2:
            if fmt[0] == 'i':
                return read_xyz(infile, **kwargs)
            elif fmt[1] == 'i':
                return read_gdat(infile, **kwargs)

    raise IOError('unrecognized file format')


def read_single(infile, fmt='auto', **kwargs):
    """Reads a single geometry from an input file.

    Unlike :func:`read_auto`, infile can be a string or open file.

    Parameters
    ----------
    infile : file or str
        The open input file or filename.
    fmt : str, optional
        The file format. Default is auto (i.e. :func:`read_auto`).
    kwargs : dict, optional
        Additional keyword arguments for the read function.

    Returns
    -------
    tuple
        The output of the read function (elem, xyz, vec, comment).
    """
    read_func = globals()['read_' + fmt]
    close = False
    infile, close = _open_file(infile, 'r')
    moldat = read_func(infile, **kwargs)
    if close:
        infile.close()

    return moldat


def read_multiple(inflist, fmt='auto', **kwargs):
    """Reads multiple files or multiple geometries into lists of data.

    Parameters
    ----------
    inflist : array_like
        The open input files or filenames.
    fmt : str, optional
        The file format. Default is auto (i.e. :func:`read_auto`).
    kwargs : dict, optional
        Additional keyword arguments for the read functions.

    Returns
    -------
    list
        The outputs of the read function for each file in inflist.

    Raises
    ------
    IOError
        When no geometries are found in the input file list.
    """
    read_func = globals()['read_' + fmt]
    inflist = np.atleast_1d(inflist)
    moldat = []
    for infile in inflist:
        infile, close = _open_file(infile, 'r')
        while True:
            try:
                moldat.append(read_func(infile, **kwargs))
            except IOError:
                break

        if close:
            infile.close()

    if len(moldat) == 0:
        raise IOError('no geometries read from input files')

    return moldat


def write_xyz(outfile, elem, xyz, vec=None, comment='', units='ang'):
    """Writes geometry to an output file in XYZ format.

    Parameters
    ----------
    outfile : file
        The open output file.
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    vec : (N, 3) array_like, optional
        The atomic cartesian vectors. If vec is None (default), it is not
        written to the output file.
    comment : str, optional
        The comment line. Default is a blank string.
    units : str, optional
        The units of the cartesian coordinates. Default is Angstroms.
    """
    natm = len(elem)
    write_xyz = xyz * con.conv('ang', units)
    outfile.write(' {}\n{}\n'.format(natm, comment))
    if vec is None:
        for atm, xyzi in zip(elem, write_xyz):
            outfile.write('{:4s}{:12.6f}{:12.6f}{:12.6f}\n'.format(atm, *xyzi))
    else:
        for atm, xyzi, pxyzi in zip(elem, write_xyz, vec):
            outfile.write('{:4s}{:12.6f}{:12.6f}{:12.6f}'.format(atm, *xyzi) +
                          '{:12.6f}{:12.6f}{:12.6f}\n'.format(*pxyzi))


def write_col(outfile, elem, xyz, vec=None, comment='', units='bohr'):
    """Writes geometry to an output file in COLUMBUS format.

    For the time being, vector output is not supported for the
    COLUMBUS file format.

    Parameters
    ----------
    outfile : file
        The open output file.
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    vec : (N, 3) array_like, optional
        The atomic cartesian vectors. If vec is None (default), it is not
        written to the output file.
    comment : str, optional
        The comment line. Default is a blank string.
    units : str, optional
        The units of the cartesian coordinates. Default is Bohr.
    """
    write_xyz = xyz * con.conv('ang', units)
    if comment != '':
        outfile.write(comment + '\n')
    for atm, (x, y, z) in zip(elem, write_xyz):
        outfile.write(' {:<2s}{:7.1f}{:14.8f}{:14.8f}{:14.8f}{:14.8f}'
                      '\n'.format(atm, con.get_num(atm), x, y, z,
                                  con.get_mass(atm)))


def write_gdat(outfile, elem, xyz, vec=None, comment='', units='bohr'):
    """Writes geometry to an output file in Geometry.dat format.

    Parameters
    ----------
    outfile : file
        The open output file.
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    vec : (N, 3) array_like, optional
        The atomic cartesian vectors. If vec is None (default), the
        vectors are replaced by zeros.
    comment : str, optional
        The comment line. Default is a blank string.
    units : str, optional
        The units of the cartesian coordinates. Default is Bohr.
    """
    natm = len(elem)
    write_xyz = xyz * con.conv('ang', units)
    outfile.write('{}\n{}\n'.format(comment, natm))
    for atm, xyzi in zip(elem, write_xyz):
        outfile.write('{:<2s}{:18.8E}{:18.8E}{:18.8E}\n'.format(atm, *xyzi))
    if vec is None:
        for line in range(natm):
            outfile.write(' {:18.8E}{:18.8E}{:18.8E}\n'.format(0, 0, 0))
    else:
        for pxyzi in vec:
            outfile.write(' {:18.8E}{:18.8E}{:18.8E}\n'.format(*pxyzi))


def write_zmt(outfile, elem, xyz, vec=None, comment='', units='ang'):
    """Writes geometry to an output file in Z-matrix format.

    At present, each atom uses the previous atoms in order as
    references. This could be made 'smarter' using the bonding module.

    For the time being, vector output is not supported for the
    Z-matrix file format.

    Parameters
    ----------
    outfile : file
        The open output file.
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    vec : (N, 3) array_like, optional
        The atomic cartesian vectors. If vec is None (default), it is not
        written to the output file.
    comment : str, optional
        The comment line. Default is a blank string.
    units : str, optional
        The units of the bond distances. Default is Angstroms.
    """
    natm = len(elem)
    if comment != '':
        outfile.write(comment + '\n')
    for i in range(natm):
        if i == 0:
            # first element has just the symbol
            outfile.write('{:s}\n'.format(elem[0]))
        elif i == 1:
            # second element has symbol, index, bond length
            outfile.write('{:<2s}{:3d}{:12.6f}'
                          '\n'.format(elem[1], 1, measure.stre(xyz, 0, 1,
                                                               units=units)))
        elif i == 2:
            # third element has symbol, index, bond length, index, bond angle
            outfile.write('{:<2s}{:3d}{:12.6f}{:3d}{:12.6f}'
                          '\n'.format(elem[2], 2, measure.stre(xyz, 1, 2,
                                                               units=units),
                                      1, measure.bend(xyz, 0, 1, 2,
                                                      units='deg')))
        else:
            # all other elements have symbol, index, bond length, index,
            # bond angle, index, dihedral angle
            outfile.write('{:<2s}{:3d}{:12.6f}{:3d}{:12.6f}{:3d}{:12.6f}'
                          '\n'.format(elem[i], i, measure.stre(xyz, i-1, i,
                                                               units=units),
                                      i-1, measure.bend(xyz, i-2, i-1, i,
                                                        units='deg'),
                                      i-2, measure.tors(xyz, i-3, i-2, i-1, i,
                                                        units='deg')))


def write_zmtvar(outfile, elem, xyz, vec=None, comment='', units='ang'):
    """Writes geometry to an output file in Z-matrix format with
    variable assignments.

    For the time being, vector output is not supported for the
    Z-matrix file format.

    Parameters
    ----------
    outfile : file
        The open output file.
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    vec : (N, 3) array_like, optional
        The atomic cartesian vectors. If vec is None (default), it is not
        written to the output file.
    comment : str, optional
        The comment line. Default is a blank string.
    units : str, optional
        The units of the bond distances. Default is Angstroms.
    """
    natm = len(elem)
    if comment != '':
        outfile.write(comment + '\n')
    vlist = []
    vdict = dict()
    for i in range(natm):
        if i == 0:
            # first element has just the symbol
            outfile.write('{:s}\n'.format(elem[0]))
        elif i == 1:
            # second element has symbol, index, bond length
            outfile.write('{:<2s}{:3d}  R{:d}'
                          '\n'.format(elem[1], 1, 1))
            vlist += ['R1']
            vdict['R1'] = measure.stre(xyz, 0, 1, units=units)
        elif i == 2:
            # third element has symbol, index, bond length, index, bond angle
            outfile.write('{:<2s}{:3d}  R{:<2d} {:3d}  A{:d}'
                          '\n'.format(elem[2], 2, 2, 1, 1))
            vlist += ['R2', 'A1']
            vdict['R2'] = measure.stre(xyz, 1, 2, units=units)
            vdict['A1'] = measure.bend(xyz, 0, 1, 2, units='deg')
        else:
            # all other elements have symbol, index, bond length, index,
            # bond angle, index, dihedral angle
            outfile.write('{:<2s}{:3d}  R{:<2d} {:3d}  A{:<2d} '
                          '{:3d}  T{:d}'
                          '\n'.format(elem[i], i, i, i-1, i-1, i-2, i-2))
            vlist += ['R'+str(i), 'A'+str(i-1), 'T'+str(i-2)]
            vdict['R'+str(i)] = measure.stre(xyz, i-1, i, units=units)
            vdict['A'+str(i-1)] = measure.bend(xyz, i-2, i-1, i, units='deg')
            vdict['T'+str(i-2)] = measure.tors(xyz, i-3, i-2, i-1, i,
                                               units='deg')
    outfile.write('\n')
    for key in vlist:
        outfile.write('{:4s} = {:14.8f}\n'.format(key, vdict[key]))


def write_traj(outfile, elem, xyz, vec=None, comment='', units='bohr',
               time=0., phase=0., ramp=0., iamp=0., state=0.):
    """Writes geometry to an output file in FMS/nomad trajectory format.

    Parameters
    ----------
    outfile : file
        The open output file.
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    vec : (N, 3) array_like, optional
        The atomic cartesian vectors. If vec is None (default), the
        vectors are written as zeros.
    comment : str, optional
        The comment line. Default is a blank string.
    units : str, optional
        The units of the cartesian coordinates. Default is Bohr.
    time : float, optional
        The time of the trajectory. Default is zero.
    phase : float, optional
        The phase of the trajectory. Default is zero.
    ramp : float, optional
        The real amplitude of the trajectory. Default is zero.
    iamp : float, optional
        The imaginary amplitude of the trajectory. Default is zero.
    state : float, optional
        The state of the trajectory. Default is zero.
    """
    natm = len(elem)
    write_xyz = xyz.flatten() * con.conv('ang', units)
    if comment != '':
        outfile.write(comment + '\n')
    if vec is None:
        write_vec = np.zeros_like(write_xyz)
    else:
        write_vec = vec.flatten()
    namp = ramp**2 + iamp**2
    args = np.hstack((write_xyz, write_vec, phase, ramp, iamp, namp, state))
    fmt = '{:10.2f}' + (6*natm + 5)*'{:10.4f}' + '\n'
    outfile.write(fmt.format(time, *args))


def write_auto(outfile, elem, xyz, vec=None, comment='', **kwargs):
    """Writes geometry to an output file based on the filename extension.

    Extensions are not case sensitive. If the extension is not recognized,
    the default format is XYZ.

    Parameters
    ----------
    outfile : file
        The open output file.
    elem : (N,) array_like
        The atomic symbols.
    xyz : (N, 3) array_like
        The atomic cartesian coordinates.
    vec : (N, 3) array_like, optional
        The atomic cartesian vectors. Default is None.
    comment : str, optional
        The comment line. Default is a blank string.
    kwargs : dict, optional
        Additional keyword arguments for the write functions.
    """
    kwargs.update(dict(vec=vec, comment=comment))
    fname = outfile.name.lower()
    ext = fname.split('.')[-1]
    if ext in ['col', 'columbus']:
        write_col(outfile, elem, xyz, **kwargs)
    elif ext == 'dat':
        write_gdat(outfile, elem, xyz, **kwargs)
    elif ext in ['zmt', 'zmat', 'zmatrix']:
        write_zmt(outfile, elem, xyz, **kwargs)
    elif ext in ['zmtvar', 'zmatvar']:
        write_zmtvar(outfile, elem, xyz, **kwargs)
    elif ext in ['tj', 'traj']:
        write_traj(outfile, elem, xyz, **kwargs)
    else:
        write_xyz(outfile, elem, xyz, **kwargs)


def write_single(outfile, moldat, fmt='auto', **kwargs):
    """Writes a single geometry to an output file.

    Unlike :func:`write_auto`, outfile can be a string or open file.

    Parameters
    ----------
    outfile : file or str
        The open output file or filename.
    moldat : tuple
        The inputs of the write function.
    fmt : str, optional
        The file format. Default is auto (i.e. :func:`write_auto`).
    kwargs : dict, optional
        Additional keyword arguments for the write functions.
    """
    write_func = globals()['write_' + fmt]
    outfile, close = _open_file(outfile, 'w')
    write_func(outfile, *moldat, **kwargs)
    if close:
        outfile.close()


def write_multiple(outfile, moldat, fmt='auto', **kwargs):
    """Writes multiple geometries into a single output file.

    Parameters
    ----------
    outfile : file or str
        The open output file or filename.
    moldat : list
        A list of the inputs of the write function for each molecule.
    fmt : str, optional
        The file format. Default is auto (i.e. :func:`write_auto`).
    kwargs : dict, optional
        Additional keyword arguments for the write functions.
    """
    moldat = np.atleast_1d(moldat)
    write_func = globals()['write_' + fmt]
    outfile, close = _open_file(outfile, 'w')
    for dat in moldat:
        write_func(outfile, *dat, **kwargs)

    if close:
        outfile.close()


def convert(inflist, outfile, infmt='auto', outfmt='auto',
            inunits=None, outunits=None, hasvec=False, hascom=False):
    """Reads a file in format infmt and writes to a file in format
    outfmt.

    Input (and output) may have multiple geometries. Z-matrix index
    ordering is not conserved.

    Parameters
    ----------
    inflist : array_like
        The open input files or filenames.
    outfile : file or str
        The open output file or filename.
    infmt : str, optional
        The input file format. Default is auto (i.e. :func:`read_auto`).
    outfmt : str, optional
        The input file format. Default is auto (i.e. :func:`write_auto`).
    inunits : str, optional
        The input distance units. If None (default), the default units
        of the read function are used.
    outunits : str, optional
        The output distance units. If None (default), the default units
        of the write function are used.
    hasvec : bool, optional
        Specifies if a vector should be read from the input file.
    hascom : bool, optional
        Specifies if a comment line should be read from the input file.
    """
    inpkw = dict(hasvec=hasvec, hascom=hascom)
    outkw = dict()
    if inunits is not None:
        inpkw.update(units=inunits)
    if outunits is not None:
        outkw.update(units=outunits)

    moldat = read_multiple(inflist, fmt=infmt, **inpkw)
    write_multiple(outfile, moldat, fmt=outfmt, **outkw)


def get_optarg(arglist, *opts, default=False):
    """Gets an optional command line argument and returns its value.

    If default is not set, the flag is treated as boolean. Note that
    that setting default to None or '' will still take an argument
    after the flag.

    Parameters
    ----------
    arglist : array_like
        The command line argument list to be parsed.
    opts : list
        The arguments searched for in arglist.
    default : str or bool
        The default value if opts are not found in arglist. If default
        is False (default), then True is returned if opts are found.

    Returns
    -------
    str or bool
        The argument value in arglist or its default value.
    """
    for op in opts:
        if op in arglist:
            ind = arglist.index(op)
            arglist.remove(op)
            if default is False:
                return True
            else:
                return arglist.pop(ind)

    return default


def _get_type(s):
    """Reads a string to see if it can be converted into int or float.

    Parameters
    ----------
    s : str
        A string to be parsed.

    Returns
    -------
    str, int or float
        The parsed value.
    """
    try:
        float(s)
        if '.' not in s:
            return 'i'
        else:
            return 'f'
    except ValueError:
        return 's'


def _valvar(unk, vardict):
    """Determines if an unknown string is a value or a dict variable.

    Parameters
    ----------
    unk : float or str
        The unknown value, either a float or a dictionary key.
    vardict : dict
        The dictionary to be searched if unk is not a float.

    Returns
    -------
    float
        The desired value for unk.

    Raises
    ------
    ValueError
        When unk is not a float and not a key in vardict.
    """
    try:
        return float(unk)
    except ValueError:
        if unk in vardict:
            return vardict[unk]
        else:
            raise KeyError('\'{}\' not found in variable list'.format(unk))


def _open_file(f, mode):
    """Opens a file if a string is provided, otherwise leaves the file open.

    Parameters
    ----------
    f : str or file
        The open file or filename.
    mode : str
        The file mode for builtin function `open` (r, w, ...).

    Returns
    -------
    file
        The desired open file.
    """
    if isinstance(f, str):
        return open(f, mode), True
    else:
        return f, False



"""
Gimbal -- Properties and transformations of 3D molecular geometries.
"""
from gimbal.molecule import import_molecule
from gimbal.molecule import import_bundle
from gimbal._version import __version__


__author__ = 'Ryan J. MacDonell'
__license__ = 'MIT'



"""
Tests for the kabsch module.
"""
import numpy as np
import gimbal.kabsch as kabsch
from examples import Geometries as eg


def test_kabsch_default():
    rot = kabsch.kabsch(eg.ch2f2[3], eg.ch2f2[1])
    soln = np.eye(3)
    soln[2,2] = -1
    assert np.allclose(rot, soln)


def test_kabsch_no_reflect():
    rot = kabsch.kabsch(eg.ch2f2[3], eg.ch2f2[1], refl=False)
    soln = -np.eye(3)
    soln[0,0] = 1
    assert np.allclose(rot, soln)


def test_map_onto_default():
    new_xyz3 = kabsch.map_onto(eg.ch2f2[0], eg.ch2f2[3], eg.ch2f2[1])
    assert np.all(np.abs(new_xyz3 - eg.ch2f2[1]) < 1e-2)


def test_map_onto_ind_cent():
    new_xyz4 = kabsch.map_onto(eg.ch2f2[0], eg.ch2f2[4], eg.ch2f2[1],
                               ind=[0,1,2], cent=0)
    assert np.allclose(new_xyz4[[0,1,2]], eg.ch2f2[1][[0,1,2]])


def test_opt_permute_default():
    new_xyz3, rmsd = kabsch.opt_permute(eg.ch2f2[0], eg.ch2f2[3], eg.ch2f2[1])
    assert np.all(np.abs(new_xyz3 - eg.ch2f2[1]) < 1e-2)
    assert rmsd < 3e-1


def test_opt_permute_plist_single():
    new_xyz3, rmsd = kabsch.opt_permute(eg.ch2f2[0], eg.ch2f2[3], eg.ch2f2[1],
                                        plist=[1,2])
    assert np.all(np.abs(new_xyz3 - eg.ch2f2[1]) < 1e-2)
    assert rmsd < 3e-3


def test_opt_permute_plist_multiple():
    new_xyz4, rmsd = kabsch.opt_permute(eg.ch2f2[0], eg.ch2f2[4], eg.ch2f2[2],
                                        plist=[[1,2],[3,4]], cent=0)
    soln = eg.ch2f2[2] / 1.02
    soln[:4] *= np.sqrt(2)/2
    soln[3] *= 1.02
    soln[4] *= 0.98
    assert np.allclose(new_xyz4, soln)


def test_opt_permute_equiv_single():
    new_xyz3, rmsd = kabsch.opt_permute(eg.ch2f2[0], eg.ch2f2[3], eg.ch2f2[1],
                                        equiv=[1,2])
    assert np.all(np.abs(new_xyz3 - eg.ch2f2[1]) < 1e-2)
    assert rmsd < 3e-3


def test_opt_permute_equiv_multiple():
    new_xyz5, rmsd = kabsch.opt_permute(eg.ch2f2[0], eg.ch2f2[5], eg.ch2f2[1],
                                        equiv=[[1,3],[2,4]], cent=0)
    soln = np.zeros((5, 3))
    soln[1,2] = 1.
    soln[2,:2] = np.sqrt(2)/2
    soln[3,0] = np.sqrt(2)
    soln[4,:2] = -np.sqrt(2)/2
    assert np.allclose(new_xyz5, soln)


def test_opt_permute_weight():
    mwgts = [12., 1., 1., 19., 19.]
    new_xyz3, rmsd = kabsch.opt_permute(eg.ch2f2[0], eg.ch2f2[3], eg.ch2f2[1],
                                        wgt=mwgts)
    assert np.all(np.abs(new_xyz3 - eg.ch2f2[1]) < 1e-2)
    assert rmsd < 5e-2


def test_opt_permute_ind():
    new_xyz5, rmsd = kabsch.opt_permute(eg.ch2f2[0], eg.ch2f2[5], eg.ch2f2[1],
                                        ind=[0,1,2], cent=0)
    assert np.allclose(new_xyz5[:3], eg.ch2f2[1][:3])
    assert rmsd < 4e-1


def test_opt_ref_default():
    new_xyz5, ind = kabsch.opt_ref(eg.ch2f2[0], eg.ch2f2[5],
                                   [eg.ch2f2[1], eg.ch2f2[2]])
    assert kabsch.rmsd(new_xyz5, eg.ch2f2[2]) < kabsch.rmsd(new_xyz5, eg.ch2f2[1])
    assert ind == 1


def test_opt_multi_default():
    match1, match2 = kabsch.opt_multi(eg.ch2f2[0], [eg.ch2f2[3], eg.ch2f2[4], eg.ch2f2[5]],
                                      [eg.ch2f2[1], eg.ch2f2[2]])
    assert kabsch.rmsd(match1[0], eg.ch2f2[1]) < kabsch.rmsd(match1[0], eg.ch2f2[2])
    assert kabsch.rmsd(match1[1], eg.ch2f2[1]) < kabsch.rmsd(match1[1], eg.ch2f2[2])
    assert kabsch.rmsd(match2[0], eg.ch2f2[2]) < kabsch.rmsd(match2[0], eg.ch2f2[1])



r"""
Module for aligning similar molecular geometries using the Kabsch algorithm.

For a reference and test set of three dimensional vectors, the Kabsch
algorithm determines the optimal rotation matrix to minimize the RMSD
between each vector pair.

After translating the vector sets of centroids, the covariance matrix **A**
is calculated by :math:`\mathbf{A} = \mathbf{P}^T \mathbf{Q}`. Then, using
singular value decomposition, :math:`\mathbf{V S W}^T = \mathbf{A}`. The
scaling component **S** is discarded. The handedness of the coordinate
system is determined by :math:`d = \mathrm{sign}(\det(\mathbf{W V}^T))`. The
rotation matrix is then found by

.. math::

    \mathbf{U} = \mathbf{W} \begin{bmatrix} 1 & 0 & 0 \\
                                            0 & 1 & 0 \\
                                            0 & 0 & d \end{bmatrix} \mathbf{V}^T

(See en.wikipedia.org/wiki/Kabsch_algorithm)

The best match for multiple references can be found by the minimum RMSD.
Sets of equivalent vectors (atoms) can be permuted as well.
"""
import itertools
import numpy as np
from scipy import linalg
import gimbal.displace as disp


def rmsd(test, ref, wgt=None):
    """Returns the root mean squared deviation of a test geometry with
    respect to a reference.

    Weights can be provided (e.g. atomic masses) with the optional
    variable wgt.

    Parameters
    ----------
    test : (N, 3) array_like
        The cartesian test geometry.
    ref : (N, 3) array_like
        The cartesian reference geometry.
    wgt : (N,) array_like, optional
        The atomic weights for computing the RMSD. If None (default),
        all weights are unity.

    Returns
    -------
    float
        The root mean squared deviation between test and ref.
    """
    if wgt is None:
        return np.sqrt(np.sum((test - ref) ** 2) / (3 * np.size(test)))
    else:
        return np.sqrt(np.sum(wgt[:,np.newaxis] * (test - ref) ** 2) /
                       (3 * np.sum(wgt) * np.size(test)))


def kabsch(test, ref, wgt=None, refl=True):
    """Returns the Kabsch rotational matrix to map a test geometry onto
    a reference.

    If weights are provided, they are used to weight the test vectors
    before forming the covariance matrix. This minimizes the weighted
    RMSD between the two geometries. If refl=True, improper rotations
    are also permitted.

    Parameters
    ----------
    test : (N, 3) array_like
        The cartesian test geometry.
    ref : (N, 3) array_like
        The cartesian reference geometry.
    wgt : (N,) array_like, optional
        The atomic weights for computing the RMSD. If None (default),
        all weights are unity.
    refl : bool, optional
        Specifies if reflections (improper rotations) are permitted.
        Default is True.

    Returns
    -------
    (3, 3) ndarray
        The rotational matrix that optimally maps test to ref.
    """
    if wgt is None:
        cov = test.T.dot(ref)
    else:
        wgt = np.array(wgt)
        cov = (wgt[:,np.newaxis] * test).T.dot(ref)
    rot1, scale, rot2 = linalg.svd(cov)
    if not refl:
        rot1[:,-1] *= linalg.det(rot1) * linalg.det(rot2)
    return rot1.dot(rot2)


def map_onto(elem, test, ref, wgt=None, ind=None, cent=None):
    """Returns the optimal mapping of a test geometry onto a reference
    using the Kabsch algorithm.

    The centre of mass of both test and ref are subtracted by default.
    If an index or list of indices is provided for cent, only the centre
    of mass of the provided atoms is subtracted. Alternatively, ind can
    be provided to only map a subset of atoms.

    Parameters
    ----------
    elem : (N,) array_like
        The atomic symbols.
    test : (N, 3) array_like
        The cartesian test geometry.
    ref : (N, 3) array_like
        The cartesian reference geometry.
    wgt : (N,) array_like, optional
        The atomic weights for computing the RMSD. If None (default),
        all weights are unity.
    ind : int or array_like, optional
        Indices of test to be mapped onto ref. If None (default), all
        atoms are used.
    cent : int or array_like, optional
        Indices used to specify the centre of mass as the origin of
        rotation. If None (default), the centre of mass is calculated
        for all atoms.

    Returns
    -------
    (N, 3) ndarray
        The test geometry optimally mapped onto the ref geometry.
    """
    if cent is None:
        cm_test = disp.get_centremass(elem, test)
        cm_ref = disp.get_centremass(elem, ref)
    else:
        cm_test = disp.get_centremass(elem[cent], test[cent])
        cm_ref = disp.get_centremass(elem[cent], ref[cent])

    new_test = test - cm_test
    new_ref = ref - cm_ref
    if ind is None:
        return new_test.dot(kabsch(new_test, new_ref, wgt=wgt)) + cm_ref
    else:
        return new_test.dot(kabsch(new_test[ind], new_ref[ind],
                                   wgt=wgt)) + cm_ref


def opt_permute(elem, test, ref, plist=None, equiv=None, wgt=None, ind=None,
                cent=None):
    """Determines optimal permutation of test geometry indices for
    mapping onto reference.

    Parameters
    ----------
    elem : (N,) array_like
        The atomic symbols.
    test : (N, 3) array_like
        The cartesian test geometry.
    ref : (N, 3) array_like
        The cartesian reference geometry.
    plist : list, optional
        A list of sets of atomic positions that can be permuted,
        parsed by :func:`_permute_elmnt`.
    equiv : array_like, optional
        A list of sets of indices that are symmetrically equivalent
        and interchangeable, parsed by :func:`_permute_group`.
    wgt : (N,) array_like, optional
        The atomic weights for computing the RMSD. If None (default),
        all weights are unity.
    ind : int or array_like, optional
        Indices of test to be mapped onto ref. If None (default), all
        atoms are used.
    cent : int or array_like, optional
        Indices used to specify the centre of mass as the origin of
        rotation. If None (default), the centre of mass is calculated

    Returns
    -------
    geom : (N, 3) ndarray
        The test geometry optimally mapped onto the ref geometry
        including atomic permutations.
    err : float
        The root mean squared deviation between test and ref.
    """
    if wgt is not None:
        wgt = np.array(wgt)
    kwargs = dict(wgt=wgt, ind=ind, cent=cent)
    if plist is None and equiv is None:
        geom = map_onto(elem, test, ref, **kwargs)
        if ind is None:
            return geom, rmsd(test, ref, wgt=wgt)
        else:
            return geom, rmsd(test[ind], ref[ind], wgt=wgt)
    else:
        eqs = _permute_group(equiv)
        prs = _permute_elmnt(plist)

        min_geom = test
        min_err = 1e10
        for i in eqs:
            for j in prs:
                xyz = np.copy(test)
                xyz[eqs[0]] = xyz[i]
                xyz[prs[0]] = xyz[j]
                geom = map_onto(elem, xyz, ref, **kwargs)
                err = rmsd(geom, ref, wgt=wgt)
                if err < min_err:
                    min_geom = geom
                    min_err = err

        return min_geom, min_err


def opt_ref(elem, test, reflist, **kwargs):
    """Determines optimal reference geometry for a given test geometry.

    Parameters
    ----------
    elem : (N,) array_like
        The atomic symbols.
    test : (N, 3) array_like
        The cartesian test geometry.
    reflist : (M, N, 3) array_like
        The cartesian reference geometries.
    kwargs : dict, optional
        Additional keyword arguments used in :func:`opt_permute`.

    Returns
    -------
    geom : (N, 3) ndarray
        The test geometry optimally mapped onto the ref geometry with
        the minimum RMSD.
    ind : int
        The index of the optimal ref geometry.
    """
    nrefs = len(reflist)
    geoms = np.empty((nrefs,) + test.shape)
    err = np.empty(nrefs)
    for i in range(nrefs):
        geoms[i], err[i] = opt_permute(elem, test, reflist[i], **kwargs)

    optref = np.argmin(err)
    return geoms[optref], optref


def opt_multi(elem, testlist, reflist, **kwargs):
    """Determines the optimal geometries of a set of test geometries
    against a set of reference geometries.

    Parameters
    ----------
    elem : (N,) array_like
        The atomic symbols.
    testlist : (L, N, 3) array_like
        The cartesian test geometries.
    reflist : (M, N, 3) array_like
        The cartesian reference geometries.
    kwargs : dict, optional
        Additional keyword arguments used in :func:`opt_permute`.

    Returns
    -------
    geomlist : (M,...) list
        The test geometries optimally mapped onto each ref geometry with
        the minimum RMSD. The L test geometries are sorted into to the
        M reference geometries.
    """
    geomlist = [[] for i in range(len(reflist))]
    for test in testlist:
        geom, ind = opt_ref(elem, test, reflist, **kwargs)
        geomlist[ind].append(geom)

    return geomlist


def _tuple2list(tupl):
    """Iteratively converts nested tuple to nested list.

    Parameters
    ----------
    tupl : tuple
        The tuple to be converted.

    Returns
    -------
    list
        The converted list of lists.
    """
    return list((_tuple2list(x) if isinstance(x, tuple) else x for x in tupl))


def _permute_elmnt(plist):
    """Generates an array of permutations of a list of lists of permutable
    indices and a list of permutable groups of indices.

    Parameters
    ----------
    plist : list
        A list of lists of indices that can be permuted.

    Returns
    -------
    list
        A list of each possible permutation of the indices.

    Examples
    --------
    >>> print(_permute_elmnt([0, 2]))
    [[0, 2], [2, 0]]
    >>> print(_permute_elmnt([[0, 2], [1, 4]]))
    [[0, 2, 1, 4], [0, 2, 4, 1], [2, 0, 1, 4], [2, 0, 4, 1]]
    """
    if plist is None:
        return [0]
    elif isinstance(plist[0], int):
        plist = [plist]

    multi_perms = [itertools.permutations(i) for i in plist]
    multi_perms = _tuple2list(itertools.product(*multi_perms))
    return [[item for sublist in i for item in sublist] for i in multi_perms]


def _permute_group(plist):
    """Generates an array of permutations of groups of indices.

    Parameters
    ----------
    plist : list
        A list of lists of indices that are symmetry equivalent.

    Returns
    -------
    list
        A list of each possible permutation of the indices.

    Examples
    --------
    >>> print(_permute_group([0, 2]))
    [[0, 2]]
    >>> print(_permute_group([[0, 2], [1, 4]]))
    [[0, 2, 1, 4], [1, 4, 0, 2]]
    """
    if plist is None:
        return [0]
    elif isinstance(plist[0], int):
        plist = [plist]

    group_perms = _tuple2list(itertools.permutations(plist))
    return [[item for sublist in i for item in sublist] for i in group_perms]



"""
Module for determining bonded atoms and natural coordinates.

The distance between N elements (atoms) in a set of cartesian coordinates
can be determined by taking the norm of the outer difference of the
coordinate array. An N x N adjacency matrix can then be formed by comparing
the distance matrix to an upper (and lower) threshold.

For an adjacency matrix **A**, the elements connected by *k* links (bonds)
is given by the matrix :math:`\mathbf{A}^k`. In molecular geometries, this
can be used to find all sets of bonds (:math:`k = 1`), angles (:math:`k = 2`)
and dihedral angles (:math:`k = 3`).

Small rings can be measured from the eigenvectors of **A**. For example, if
:math:`\mathbf{B} = \mathrm{eig}(\mathbf{A})`, the number of three-membered
rings is given by :math:`\sum_{ij} \mathbf{B}_{ij}^3 / 6`.
"""
import numpy as np
from scipy import linalg
import gimbal.constants as con


def build_adjmat(elem, xyz, error=0.56, lothresh=0.4):
    """Returns an adjacency matrix from a set of atoms.

    At present, thresholds are set to the Rasmol defaults of covalent
    radius + 0.56 and covalent radius - 0.91.

    Parameters
    ----------
    elem : (N,) array_like
        List of atomic symbols.
    xyz : (N, 3) array_like
        Array of cartesian atomic positions.
    error : float, optional
        The allowable error for the upper bound of bond lengths in
        Angstroms. Default is 0.56 based on JMol.
    lothresh : float, optional
        The lower bound for all bond lengths. Default is 0.4 based on JMol.

    Returns
    -------
    (N, N) ndarray
        The adjacency matrix, an N x N array of ones at off-diagonal
        elements where two atoms are bonded, zeros otherwise.
    """
    rad = con.get_covrad(elem)
    upthresh = np.add.outer(rad, rad) + error

    xyz_diff = xyz.T[:,:,np.newaxis] - xyz.T[:,np.newaxis,:]
    blength = np.sqrt(np.sum(xyz_diff**2, axis=0))

    bonded = (blength < upthresh) & (blength > lothresh)
    return bonded.astype(int)


def power(mat, k):
    """Returns the kth power of a square matrix.

    The elements :math:`(\mathbf{A}^k)_{ij}` of the *k*-th power of an
    adjacency matrix represent the number of *k*-length paths from element
    *i* to element *j*, including repetitions.

    Parameters
    ----------
    mat : (N, N) array_like
        The input square matrix, **A**.
    k : int
        The power of the matrix.

    Returns
    -------
    (N, N) ndarray
        The input matrix to the power k, A^k.
    """
    return np.linalg.matrix_power(mat, k)


def path_len(adjmat, k):
    """Returns the matrix of paths of length k from an adjacency matrix.

    Ideally, all elements should be unity unless loops are present. Loops
    are not fully accounted for at the moment. They should lead to
    nonzero diagonal elements.

    Parameters
    ----------
    adjmat : (N, N) array_like
        The adjacency matrix.
    k : int
        The desired path length.

    Returns
    -------
    (N, N) ndarray
        The matrix of elements connected by k bonds.
    """
    new_mat = power(adjmat, k)
    new_mat -= np.diag(np.diag(new_mat))
    return new_mat


def num_neighbours(adjmat, k):
    """Returns the number of atoms k atoms away from each atom.

    Parameters
    ----------
    adjmat : (N, N) array_like
        The adjacency matrix.
    k : int
        The desired path length

    Returns
    -------
    (N,) ndarray
        A vector of length N with the number of path of length k for
        each atom.
    """
    return np.sum(path_len(adjmat, k), axis=0)


def num_loops(adjmat, k):
    """Returns the number of loops of length k.

    Only works for 3-loops and 4-loops at the moment.

    Parameters
    ----------
    adjmat : (N, N) array_like
        The adjacency matrix.
    k : int
        The desired number of atoms in the loop.

    Returns
    -------
    int
        The number of loops with k atoms.
    """
    if k < 3:
        raise ValueError('Loops must have 3 or more elements.')

    eigs = linalg.eigh(adjmat)[0]
    if k == 3:
        loop = np.sum(eigs ** 3) / 6
    elif k == 4:
        adj2 = power(adjmat, 2)
        loop = (np.sum(eigs ** 4) - 2 * np.sum(adj2) + np.sum(adjmat)) / 8
    else:
        raise ValueError('Loops of more than 4 elements not supported.')

    return int(round(loop))



"""
Tests for the constants module
"""
import pytest
import numpy as np
import gimbal.constants as con


def test_symbol_lookup():
    assert con.sym[0] == 'X'


def test_mass_lookup():
    assert np.isclose(con.mass[0], 0.)


def test_covrad_lookup():
    assert np.isclose(con.covrad[0], 0.)


def test_lenunits_lookup():
    assert np.isclose(con.lenunits['auto'], 1.)


def test_angunits_lookup():
    assert np.isclose(con.angunits['auto'], 1.)


def test_timunits_lookup():
    assert np.isclose(con.timunits['auto'], 1.)


def test_masunits_lookup():
    assert np.isclose(con.masunits['auto'], 1.)


def test_eneunits_lookup():
    assert np.isclose(con.eneunits['auto'], 1.)


def test_get_num_single():
    assert con.get_num('H') == 1


def test_get_num_deuterium():
    assert con.get_num('D') == 1


def test_get_num_list():
    assert np.all(con.get_num(['H', 'He', 'Li']) == np.array([1, 2, 3]))


def test_get_num_fails():
    with pytest.raises(ValueError, match=r'Unrecognized atomic symbol .*'):
        con.get_num(['J', 'L', 'M'])


def test_get_mass():
    assert np.isclose(con.get_mass('H'), 1.00782504)


def test_get_covrad():
    assert np.isclose(con.get_covrad('H'), 0.320)


def test_unit_vec():
    vec_len = np.linalg.norm(con.unit_vec([1., -1., 2.]))
    assert np.isclose(vec_len, 1.)


def test_unit_vec_fails():
    with pytest.raises(ValueError, match=r'Cannot make unit vector from .*'):
        uvec = con.unit_vec(np.zeros(3))


def test_arccos_plusone():
    ang = con.arccos(1 + 1e-10)
    assert np.isclose(ang, 0)


def test_arccos_minusone():
    ang = con.arccos(-1 - 1e-10)
    assert np.isclose(ang, np.pi)


def test_conv_unit():
    assert np.isclose(con.conv('auto', 'auto'), 1.)


def test_conv_len():
    assert np.isclose(con.conv('ang', 'pm'), 100.)


def test_conv_ang():
    assert np.isclose(con.conv('deg', 'rad'), np.pi/180.)


def test_conv_tim():
    assert np.isclose(con.conv('ps', 'fs'), 1e3)


def test_conv_mas():
    assert np.isclose(con.conv('me', 'mp'), 1836.15267981)


def test_conv_ene():
    assert np.isclose(con.conv('har', 'ev'), 27.21138505)


def test_conv_fails():
    with pytest.raises(ValueError, match=r'.* not of same unit type'):
        con.conv('ev', 'fs')



"""
Setup script for the Gimbal package
"""
from setuptools import setup
from setuptools import find_packages


def readme():
    """Returns the contents of the README without the header image."""
    header = '======\nGimbal\n======\n'
    with open('README.rst', 'r') as f:
        f.readline()
        return header + f.read()


def requirements():
    """Returns the requirement list."""
    with open('requirements.txt', 'r') as f:
        return [line.strip() for line in f.readlines()]


# read the current version number
exec(open('gimbal/_version.py').read())


setup(
    name='gimbal',
    version=__version__,
    description=('Tools for importing, creating, editing and querying ' +
                 'molecular geometries'),
    long_description=readme(),
    long_description_content_type='text/x-rst',
    keywords='gimbal molecule geometry displacement transformation 3D',
    url='https://github.com/ryjmacdonell/gimbal',
    author='Ryan J. MacDonell',
    author_email='rmacd054@uottawa.ca',
    license='MIT',
    packages=find_packages(),
    scripts=['bin/convgeom', 'bin/measure', 'bin/nudge', 'bin/subst'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Chemistry'
                 ],
    install_requires=requirements()
      )



"""
Constant molecular properties to be used in other modules.

Arrays are ordered by atomic number for convenience. Atomic symbols are case
sensitive. This module should not depend on other modules.

Unit types
----------
Length: Angstrom (ang), Bohr (bohr), picometre (pm), nanometre (nm)

Angle: radian (rad), degree (deg)

Time: femtosecond (fs), picosecond (ps), atomic unit (au)

Mass: atomic mass unit (amu), electron mass (me), proton mass (mp),
kilogram (kg)

Energy: electron volt (ev), Hartree (har), kilocalorie per mole (kcm),
kilojoule per mole (kjm), reciprocal centimetre (cm)

For all types, 'auto' will give the default unit.

Attributes
----------
sym : ndarray
    List of atomic symbols up to Krypton. The ordering (with the
    exception of deuterium) yields the correct atomic number from
    ``sym.index(elem)``.
mass : ndarray
    List of atomic masses corresponding to the elements in `sym`.
covrad : ndarray
    List of covalent radii corresponding to the elements in `sym`.
lenunits : dict
    Dictionary of units of length and their conversions from the
    default (angstroms).
angunits : dict
    Dictionary of units of angle and their conversions from the
    default (radians).
timunits : dict
    Dictionary of units of time and their conversions from the
    default (femtoseconds).
masunits : dict
    Dictionary of units of mass and their conversions from the
    default (atomic mass units).
eneunits : dict
    Dictionary of units of energy and their conversion from the
    default (electron volts).
"""
import numpy as np


# Global constants
sym = np.array(['X', 'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne',
                'Na', 'Mg', 'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca',
                'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn',
                'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr'])
mass = np.array([0.00000000, 1.00782504, 4.00260325, 7.01600450, 9.01218250,
                 11.00930530, 12.00000000, 14.00307401, 15.99491464,
                 18.99840325, 19.99243910, 22.98976970, 23.98504500,
                 26.98154130, 27.97692840, 30.97376340, 31.97207180,
                 34.96885273, 39.96238310, 38.96370790, 39.0983,
                 40.078, 44.955908, 47.867, 50.9415, 51.9961, 54.938044,
                 55.845, 58.933194, 58.6934, 63.546, 65.38, 69.723, 72.630,
                 74.921595, 78.971, 79.904, 83.798])
covrad = np.array([0.000, 0.320, 1.600, 0.680, 0.352, 0.832, 0.720, 0.680,
                   0.680, 0.640, 1.120, 0.972, 1.100, 1.352, 1.200, 1.036,
                   1.020, 1.000, 1.568, 1.328, 0.992, 1.440, 1.472, 1.328,
                   1.352, 1.352, 1.340, 1.328, 1.620, 1.520, 1.448, 1.220,
                   1.168, 1.208, 1.220, 1.208, 1.600])
lenunits = dict(auto=1., ang=1., bohr=1./0.52917721, pm=100., nm=0.1)
angunits = dict(auto=1., rad=1., deg=180./np.pi)
timunits = dict(auto=1., fs=1., ps=0.001, au=0.024188843)
masunits = dict(auto=1., amu=1., me=5.48579909e-4, mp=1.00727647,
                kg=1.66053904e-27)
eneunits = dict(auto=1., ev=1., har=1./27.21138505, kcm=23.061, kjm=96.485,
                cm=8065.5)


def get_num(elem):
    """Returns atomic number from atomic symbol.

    Takes advantage of the fact that sym indices match atomic numbers.

    Parameters
    ----------
    elem : str or array_like
        The atomic symbol(s) to be parsed.

    Returns
    -------
    int or ndarray
        The atomic numbers corresponding to each symbol.
    """
    if isinstance(elem, str):
        return _find_index(elem)
    else:
        for atm in elem:
            if atm not in sym and atm[0] not in ['X', 'D']:
                raise ValueError('Unrecognized atomic symbol \'' + atm +
                                 '\'. Use X prefix for dummy atoms.')
        return np.array([_find_index(atm) for atm in elem])


def get_mass(elem):
    """Returns atomic mass from atomic symbol.

    Parameters
    ----------
    elem : str of array_like
        The atomic symbol(s) to be parsed.

    Returns
    -------
    float or ndarray
        The atomic masses corresponding to each symbol.
    """
    return mass[get_num(elem)]


def get_covrad(elem):
    """Returns covalent radius from atomic symbol.

    Parameters
    ----------
    elem : str of array_like
        The atomic symbol(s) to be parsed.

    Returns
    -------
    float or ndarray
        The atomic covalent radii corresponding to each symbol.
    """
    return covrad[get_num(elem)]


def unit_vec(v):
    """Returns a unit vector aligned with a given vector.

    Parameters
    ---------
    v : array_like
        The input, un-normalized vector.

    Returns
    -------
    ndarray
        The normalized (unit) vector."""
    vlen = np.linalg.norm(v)
    if np.isclose(vlen, 0):
        raise ValueError('Cannot make unit vector from zero vector.')
    else:
        return v / vlen


def arccos(val):
    """Returns the arccosine of an angle allowing for numerical errors.

    NumPy's arccos function is defined for the range [-1, 1], but
    returns NaN for :math:`|x| = 1 + \delta`, where :math:`\delta` is
    small.  This can be avoided by checking for limiting cases with
    ``numpy.isclose``.

    Parameters
    ----------
    val : float
        The x-coordinate on the unit circle.

    Returns
    -------
    float
        The angle intersecting the unit circle at x = val.
    """
    if np.isclose(val, -1):
        return np.pi
    elif np.isclose(val, 1):
        return 0.
    else:
        return np.arccos(val)


def conv(old='auto', new='auto'):
    """Returns conversion factor from old units to new units.

    Parameters
    ----------
    old : str, optional
        The units to be converted from. See different units types for
        defaults.
    new : str, optional
        The units to be converted to. See different units types for
        defaults.

    Returns
    -------
    float
        The conversion factor, new_units / old_units.
    """
    if old == new:
        return 1.
    for unittype in [lenunits, angunits, timunits, masunits, eneunits]:
        if old in unittype and new in unittype:
            return unittype[new] / unittype[old]

    raise ValueError('Units \'{}\' and \'{}\' unrecognized or '
                     'not of same unit type'.format(old, new))


def _find_index(string):
    """Determines if dummy or regular atom and returns index.

    Parameters
    ----------
    string : str
        The atomic symbol.

    Returns
    -------
    int
        The atomic number of the given atomic symbol.
    """
    if string[0] == 'X':
        return 0
    elif string  == 'D':
        return 1
    else:
        return np.where(sym == string)[0][0]



"""
Tests for the symmetry module.
"""
import numpy as np
import gimbal.symmetry as symmetry


def test_principle_axes():
    pass


def test_assign_rotor():
    pass


def test_symmetrize():
    pass


def test_cart_jumble():
    pass



"""
Tests for the bonding module.
"""
import pytest
import numpy as np
import gimbal.bonding as bonding
from examples import Geometries as eg


def test_build_adjmat_default():
    amat = bonding.build_adjmat(*eg.h2)
    assert np.all(amat == np.array([[0, 1], [1, 0]]))


def test_build_adjmat_error():
    amat = bonding.build_adjmat(*eg.h2, error=0.)
    assert np.all(amat == 0)


def test_build_adjmat_lothresh():
    amat = bonding.build_adjmat(*eg.h2, lothresh=0.8)
    assert np.all(amat == 0)


def test_power():
    mat = np.array([[1, -1], [1, 1]])
    prod = np.array([[-2, -2], [2, -2]])
    assert np.all(bonding.power(mat, 3) == prod)


def test_path_len():
    amat = bonding.build_adjmat(*eg.ch4)
    soln = np.ones(5, dtype=int) - np.eye(5, dtype=int)
    soln[0,:] = 0
    soln[:,0] = 0
    assert np.all(bonding.path_len(amat, 2) == soln)


def test_num_neighbours():
    amat = bonding.build_adjmat(*eg.ch4)
    soln = np.array([0, 3, 3, 3, 3])
    assert np.all(bonding.num_neighbours(amat, 2) == soln)


def test_num_loops_3():
    amat = bonding.build_adjmat(*eg.p4)
    assert bonding.num_loops(amat, 3) == 4


def test_num_loops_4():
    amat = bonding.build_adjmat(*eg.c4h4)
    assert bonding.num_loops(amat, 4) == 1


def test_num_loops_less_than_3():
    amat = bonding.build_adjmat(*eg.ch4)
    with pytest.raises(ValueError, match=r'Loops must have 3 .*'):
        bonding.num_loops(amat, 2)


def test_num_loops_more_than_3():
    amat = bonding.build_adjmat(*eg.ch4)
    with pytest.raises(ValueError, match=r'Loops of more than 4 .*'):
        bonding.num_loops(amat, 5)



"""
Tests for the displace module.
"""
import pytest
import numpy as np
import gimbal.displace as displace
from examples import Geometries as eg


def test_VectorParser_explicit():
    vp = displace.VectorParser()
    assert np.allclose(vp('[1., 2., 3.]'), [1, 2, 3])


def test_VectorParser_vector():
    vp = displace.VectorParser()
    assert np.allclose(vp([1, 2, 3]), [1, 2, 3])


def test_VectorParser_cross_unit():
    vp = displace.VectorParser()
    ax = np.array([[6, -3, 2], [1, 1, 1], [2, 2, 0]])
    ax2 = np.array([1, 2, 3])
    assert np.allclose(vp(ax, unit=True), ax2 / np.linalg.norm(ax2))


def test_VectorParser_xyz_getter():
    xyz = eg.c2h4[1]
    vp = displace.VectorParser(xyz)
    assert np.allclose(vp.xyz, xyz)


def test_VectorParser_xyz_unary():
    xyz = eg.c2h4[1]
    vp = displace.VectorParser(xyz)
    assert np.allclose(vp('-1'), -xyz[1])


def test_VectorParser_xyz_subtract():
    xyz = eg.c2h4[1]
    vp = displace.VectorParser(xyz)
    assert np.allclose(vp('3 - 1'), xyz[3] - xyz[1])


def test_VectorParser_xyz_cross():
    xyz = eg.c2h4[1]
    vp = displace.VectorParser(xyz)
    soln = np.cross(xyz[3] - xyz[1], xyz[4] - xyz[1])
    assert np.allclose(vp('(3 - 1) x (4 - 1)'), soln)


def test_VectorParser_xyz_proj():
    xyz = eg.c2h4[1]
    vp = displace.VectorParser(xyz)
    soln = np.dot(xyz[3], xyz[1]) / np.dot(xyz[1], xyz[1]) * xyz[1]
    assert np.allclose(vp('(3 o 1) / (1 o 1) * 1'), soln)


def test_VectorParser_xyz_mixed():
    xyz = eg.c2h4[1]
    vp = displace.VectorParser(xyz)
    soln = [1.2, xyz[3,0]**2, np.dot(xyz[3], xyz[2]-xyz[1])]
    assert np.allclose(vp('[1.2, 3[0]^2., 3 o (2 - 1)]'), soln)


def test_VectorParser_unrec_axis():
    vp = displace.VectorParser()
    with pytest.raises(ValueError, match=r'Axis specification not recognized'):
        rot = vp([1, 1, 1, 1])


def test_translate_ax_all():
    new_xyz = displace.translate(eg.c2h4[1], np.sqrt(3), [1, 1, 1])
    diff_xyz = new_xyz - eg.c2h4[1]
    assert np.allclose(diff_xyz, np.ones((6, 3)))


def test_translate_x_carbon():
    new_xyz = displace.translate(eg.c2h4[1], 1., 'X', ind=[0,1])
    diff_xyz = new_xyz - eg.c2h4[1]
    soln = np.zeros((6, 3))
    soln[:2,0] = 1.
    assert np.allclose(diff_xyz, soln)


def test_translate_x_bohr():
    new_xyz = displace.translate(eg.c2h4[1], 2., '-X', units='bohr')
    diff_xyz = new_xyz - eg.c2h4[1]
    soln = np.zeros((6, 3))
    soln[:,0] = -1.05835442
    assert np.allclose(diff_xyz, soln)


def test_rotmat_zero():
    rot = displace.rotmat(0., [0, 0, 1])
    assert np.allclose(rot, np.eye(3))


def test_rotmat_invert():
    rot = displace.rotmat(np.pi, [0, 0, 1], det=-1)
    assert np.allclose(rot, -np.eye(3))


def test_rotmat_degrees():
    rot = displace.rotmat(90., [0, 1, 0], units='deg')
    soln = np.flip(np.eye(3), axis=0)
    soln[0,2] = -1.
    assert np.allclose(rot, soln)


def test_rotmat_det_error():
    with pytest.raises(ValueError, match=r'Determinant of a rotational .*'):
        rot = displace.rotmat(0., [0, 0, 1], det=2.)


def test_angax_identity():
    ang, ax, d = displace.angax(np.eye(3))
    assert np.isclose(ang, 0.)
    assert np.allclose(ax, [0, 0, 1])
    assert np.isclose(d, 1.)


def test_angax_invert():
    ang, ax, d = displace.angax(-np.eye(3))
    assert np.isclose(ang, np.pi)
    assert np.allclose(ax, [0, 0, 1])
    assert np.isclose(d, -1.)


def test_angax_pi():
    rot = -np.eye(3)
    rot[1,1] = 1.
    ang, ax, d = displace.angax(rot)
    assert np.isclose(ang, np.pi)
    assert np.allclose(ax, [0, 1, 0])
    assert np.isclose(d, 1.)


def test_angax_degrees():
    u = np.ones(3) / np.sqrt(3)
    rot = np.array([[0, u[2], -u[1]], [-u[2], 0, u[0]], [u[1], -u[0], 0]])
    rot += np.outer(u, u)
    ang, ax, d = displace.angax(rot, units='deg')
    assert np.isclose(ang, 90.)
    assert np.allclose(ax, u)
    assert np.isclose(d, 1.)


def test_angax_det_error():
    with pytest.raises(ValueError, match=r'Determinant of a rotational .*'):
        ang, ax, d = displace.angax(2*np.eye(3))


def test_rotate_degrees():
    xyz = eg.c2h4[1]
    new_xyz = displace.rotate(xyz, 90., 'Z', units='deg')
    soln = np.array([-xyz[:,1], -xyz[:,0], xyz[:,2]]).T
    assert np.allclose(new_xyz, soln)


def test_rotate_invert():
    new_xyz = displace.rotate(eg.c2h4[1], np.pi, '-Z', det=-1)
    assert np.allclose(new_xyz, -eg.c2h4[1])


def test_rotate_reflect():
    soln = np.copy(eg.c2h4[1])
    new_xyz = displace.rotate(soln, 0., 'XxZ', det=-1)
    soln[:,1] *= -1
    assert np.allclose(new_xyz, soln)


def test_rotate_carbons():
    soln = np.copy(eg.c2h4[1])
    new_xyz = displace.rotate(soln, np.pi, 'X', ind=[0,1])
    soln[[0, 1]] = soln[[1, 0]]
    assert np.allclose(new_xyz, soln)


def test_rotate_origin():
    xyz = np.copy(eg.c2h4[1])
    new_xyz = displace.rotate(xyz, np.pi/2, 'Y', origin=xyz[0])
    soln = np.array([xyz[:,2] - xyz[0,2], xyz[:,1], xyz[0,2]*np.ones(6)]).T
    assert np.allclose(new_xyz, soln)


def test_align_pos_default():
    xyz = np.ones((1, 3))
    new_xyz = displace.align_pos(xyz, -np.ones(3), np.zeros(3))
    assert np.allclose(new_xyz, 2*np.ones(3))


def test_align_pos_ind():
    xyz = np.array([np.ones(3), np.zeros(3)])
    new_xyz = displace.align_pos(xyz, xyz[0], xyz[1], ind=0)
    assert np.allclose(new_xyz[0], new_xyz[1])


def test_align_axis_default():
    xyz = np.copy(eg.c2h4[1])
    new_xyz = displace.align_axis(xyz, xyz[0]-xyz[1], 'Y')
    soln = np.array([xyz[:,0], xyz[:,2], -xyz[:,1]]).T
    assert np.allclose(new_xyz, soln)


def test_align_axis_same():
    ax1 = np.array([1., -2., 0.])
    ax2 = 2*ax1
    xyz = np.copy(eg.c2h4[1])
    new_xyz = displace.align_axis(xyz, ax1, ax2)
    assert np.allclose(new_xyz, xyz)


def test_align_axis_pi_y():
    ax = np.array([0., -2., 0.])
    soln = np.copy(eg.c2h4[1])
    new_xyz = displace.align_axis(soln, ax, 'Y')
    soln[:,1] *= -1
    assert np.allclose(new_xyz, soln)


def test_align_axis_pi_z():
    ax = np.array([0., 0., -2.])
    soln = np.copy(eg.c2h4[1])
    new_xyz = displace.align_axis(soln, ax, 'Z')
    soln[:,2] *= -1
    assert np.allclose(new_xyz, soln)


def test_align_axis_ind():
    soln = np.copy(eg.c2h4[1])
    new_xyz = displace.align_axis(soln, 'Z', '-Y', ind=[0,1])
    soln[[0,1],1] = -soln[[0,1],2]
    soln[[0,1],2] = 0.
    assert np.allclose(new_xyz, soln)


def test_align_axis_origin():
    xyz = np.copy(eg.c2h4[1])
    new_xyz = displace.align_axis(xyz, 'Z', 'X', origin=xyz[0])
    soln = np.array([xyz[:,2] - xyz[0,2], xyz[:,1], xyz[0,2]*np.ones(6)]).T
    assert np.allclose(new_xyz, soln)


def test_get_centremass_molecule():
    xyz = eg.c2h4[1] + np.ones(3)
    cm = displace.get_centremass(eg.c2h4[0], xyz)
    assert np.allclose(cm, np.ones(3))


def test_get_centremass_atom():
    cm = displace.get_centremass(eg.c2h4[0][0], eg.c2h4[1][0])
    assert np.allclose(cm, eg.c2h4[1][0])


def test_centre_mass():
    xyz = eg.c2h4[1] + np.ones(3)
    new_xyz = displace.centre_mass(eg.c2h4[0], xyz)
    diff_xyz = new_xyz - xyz
    assert np.allclose(diff_xyz, -np.ones((6, 3)))
